

#include <vector>
#include "blas_enum.h"
#include "nist_spblas.h"
#include <iostream>

using namespace std;

namespace gov_nist_math_SPBLAS
{

static vector<Sp_mat *> Table;
static unsigned int Table_active_matrices = 0;
int Table_insert(Sp_mat* S);
int Table_remove(int i);

/* 
  finds an empty slot in global sparse marix table, and fills it with
  given entry.  Returns -1 if no spot found, or table is corrupt.
*/
int Table_insert(Sp_mat* S)
{
  if (Table_active_matrices <= Table.size())
  {
    Table.push_back(S);
    Table_active_matrices++;
    return Table.size() - 1;
  }
  else
  {
    /* there is an available slot; find it. */
    for (unsigned int i=0; i<Table.size(); i++)
    {
      if (Table[i] == NULL)
      {
        Table[i] = S;
        Table_active_matrices++;
        return i;
      }
    }
  }

  return -1;
}

/* 
  removes an exisiting sparse matrix from global table.  Returns 0, if
  successfull, 1 otherwise.
*/
int Table_remove(unsigned int i)
{
  if (i < Table.size() && Table[i] != NULL)
  {
    Table[i] = NULL;
    Table_active_matrices--;
    return 0;
  }
  else
    return -1;
}


void Sp_mat::print() const
{
  

  cout << "State : " <<
    (is_void() ? "void" : 
     is_new()  ? "new" :  
     is_open() ? "open" :
     is_valid() ? "valid" : "unknown") << "\n";

  cout << "M = " <<  num_rows() <<  "  N = " << num_cols() <<
        "  nz = " << num_nonzeros() << "\n";

#define yesno(exp) ( (exp) ? "yes" : "no" )

  cout << "real: "     << yesno(is_real()) << "\n";
  cout << "complex: "  << yesno(is_complex()) << "\n";
  cout << "double "    << yesno(is_double_precision()) << "\n";
  cout << "single "    << yesno(is_single_precision()) << "\n";

  cout << "upper_triangular: " << yesno(is_upper_triangular()) << "\n";
  cout << "lower_triangular: " << yesno(is_lower_triangular()) << "\n";
      
#undef yesno
}

void table_print()
{
  cout << "Table has " << Table.size() << " element(s). \n";
  for (unsigned int i=0; i< Table.size(); i++)
  {
    if (Table[i] != 0)
    {
      cout << "***** Table[" << i << "]: \n";
      Table[i]->print();
      cout << "\n\n";
    }
  }
}

void print(int A)
{
  cout << "\n";
  Table[A]->print();
  cout << "\n";
}

}
/* namespace NIST */

using namespace std;


using namespace gov_nist_math_SPBLAS;


/* Level 1 */


/* dummy routines for real version of usdot to compile. */

const double conj(const double &x)
{ 
  return x;
}

const float conj(const float &x)
{ 
  return x;
}


void BLAS_dusdot( enum blas_conj_type conj_flag, int nz,
    const double *x,  const int *index,  const double *y, int incy,
    double *r, enum blas_base_type index_base)
{
   BLAS_xusdot(conj_flag, nz, x, index, y, incy, r, index_base);
}


void BLAS_dusaxpy(int nz, double alpha, const double *x,
    const int *index, double *y, int incy,
    enum blas_base_type index_base)
{
  BLAS_xusaxpy(nz, alpha, x, index, y, incy, index_base);
}

void BLAS_dusga( int nz, const double *y, int incy, double *x, const int *indx,
              enum blas_base_type index_base )
{
  BLAS_xusga( nz, y, incy, x, indx, index_base);

}

void BLAS_dusgz( int nz, double *y, int incy, double *x, const int *indx,
              enum blas_base_type index_base )
{
  BLAS_xusgz(nz, y, incy, x, indx, index_base);
}


void BLAS_dussc(int nz, const double *x, double *y, int incy, const int *index,
    enum blas_base_type index_base)
{
  BLAS_xussc(nz, x, y, incy, index, index_base);
}



/* Level 2/3 */

int BLAS_duscr_begin(int M, int N)
{
  DSp_mat *A = new DSp_mat(M, N);
  A->set_double_precision();
  A->set_real();

  return Table_insert(A);
}

int BLAS_fuscr_begin(int M, int N)
{
  FSp_mat *A = new FSp_mat(M, N);
  A->set_single_precision();
  A->set_real();

  return Table_insert(A);
}

int BLAS_cuscr_begin(int M, int N)
{
  CSp_mat *A = new CSp_mat(M, N);
  A->set_single_precision();
  A->set_complex();

  return Table_insert(A);
  
}

int BLAS_zuscr_begin(int M, int N)
{
  ZSp_mat *A = new ZSp_mat(M, N);
  A->set_double_precision();
  A->set_complex();

  return Table_insert(A);

}




int BLAS_fuscr_end(int A)
{
  return ((TSp_mat<float>*) Table[A])->end_construction();
}

int BLAS_duscr_end(int A)
{
  return ((TSp_mat<double>*) Table[A])->end_construction();
}

int BLAS_cuscr_end(int A)
{
  return ((TSp_mat<complex<float> >*) Table[A])->end_construction();
}

int BLAS_zuscr_end(int A)
{
  return ((TSp_mat<complex<double> >*) Table[A])->end_construction();
}



/* USCR_INSERT_ENTRY() */

template <class T>
int BLAS_xuscr_insert_entry(int A,  T val, int i, int j)
{
  return ((TSp_mat<T> *)Table[A])->insert_entry(val, i, j);

}

int BLAS_fuscr_insert_entry(int A, float Val, int i, int j)
{
  return BLAS_xuscr_insert_entry(A, Val, i, j);
}

int BLAS_duscr_insert_entry(int A, double Val, int i, int j)
{
  return BLAS_xuscr_insert_entry(A, Val, i, j);
}

int BLAS_cuscr_insert_entry(int A, void *Val, int i, int j)
{

  return BLAS_xuscr_insert_entry(A, *(complex<float> *)Val, i, j);
}


/* --------------------- */
/* USCR_INSERT_ENTRIES() */
/* --------------------- */

template <class T> 
int BLAS_xuscr_insert_entries(int A, int nz, const T* Val, const int* I, 
      const int *J)
{
  return ((TSp_mat<T>*) Table[A])->insert_entries(nz, Val, I, J);
}

int BLAS_fuscr_insert_entries(int A, int nz, const float *Val, const int* I, 
    const int* J)
{
  return BLAS_xuscr_insert_entries(A, nz, Val, I, J);
}


int BLAS_duscr_insert_entries(int A, int nz, const double *Val, const int* I, 
    const int* J)
{
  return BLAS_xuscr_insert_entries(A, nz, Val, I, J);
}


int BLAS_cuscr_insert_entries(int A, int nz, const void *Val, 
    const int* I, const int* J)
{
  return BLAS_xuscr_insert_entries(A, nz, (const complex<float> *)Val, I, J);
}

int BLAS_zuscr_insert_entries(int A, int nz, const void *Val, 
    const int* I, const int* J)
{
  return BLAS_xuscr_insert_entries(A, nz, (const complex<double> *)Val, I, J);
}



/* --------- */
/*   USMV()  */ 
/* --------- */

int BLAS_fusmv(enum blas_trans_type transa, float alpha, int A, 
  const float *x, int incx, float *y, int incy )
{
  return BLAS_xusmv(transa,   alpha, A, x, incx, y, incy);
}


int BLAS_dusmv(enum blas_trans_type transa, double alpha, int A, 
  const double *x, int incx, double *y, int incy )
{
  return BLAS_xusmv(transa,   alpha, A, x, incx, y, incy);
}


int BLAS_cusmv(enum blas_trans_type transa, const void * alpha, int A, 
  const void *x, int incx, void *y, int incy )
{
//  complex<float> alpha_c = * ((complex<float> *) alpha);

  return BLAS_xusmv(transa,  * ((complex<float> *) alpha), A, 
      (const complex<float> *)x, incx, 
      (      complex<float> *)y, incy);
}


int BLAS_zusmv(enum blas_trans_type transa, const void * alpha, int A, 
  const void *x, int incx, void *y, int incy )
{
//  complex<double> alpha_c = * ((complex<double> *) alpha);

  return BLAS_xusmv(transa,  * ((complex<double> *) alpha), A, 
      (const complex<double> *)x, incx, 
      (      complex<double> *)y, incy);
}



int BLAS_dussv( enum blas_trans_type transa, double alpha, 
    int A, double *x, int incx )
{
  return BLAS_xussv( transa, alpha, A, x, incx);
}


/* -------- */
/*  USSP()  */
/* -------- */
int BLAS_ussp(int A, int pname)
{
    Sp_mat *S = Table[A];


                                /* Note: these are returns, in the case */
                                /* statement, so "break" is not needed.  */
    switch (pname)
    {
        case (blas_unit_diag) : S->set_unit_diag(); break;
        case (blas_complex) :   S->set_complex(); break;
        case (blas_real)    :   S->set_real(); break;

        case (blas_single_precision) : S->set_single_precision(); break;
        case (blas_double_precision) : S->set_double_precision(); break;


        case (blas_lower_triangular) : S->set_lower_triangular(); break;
        case (blas_upper_triangular) : S->set_upper_triangular(); break;

        case (blas_lower_symmetric) : S->set_lower_symmetric(); break;
        case (blas_upper_symmetric) : S->set_upper_symmetric(); break;

        case (blas_lower_hermitian) : S->set_lower_hermitian(); break;
        case (blas_upper_hermitian) : S->set_upper_hermitian(); break;


                                        /* optimizations not used */
        case (blas_regular )          :
        case (blas_irregular)       : 
        case (blas_block)           :
        case (blas_unassembled)     :    return 0;

        default:                return -1;  /* invalid property */
    }

    return 0;

}

/* -------- */
/*  USSP()  */
/* -------- */

int BLAS_usgp(int A, int pname)
{
  Sp_mat *S = Table[A];

    switch (pname)
    {
        case (blas_unit_diag) : return S->is_unit_diag(); 

        case (blas_num_rows)          : return S->num_rows(); 
        case (blas_num_cols)          : return S->num_cols(); 
        case (blas_num_nonzeros)      : return S->num_nonzeros(); 

        case (blas_single_precision)  : return S->is_single_precision();
        case (blas_double_precision)  : return S->is_double_precision();


        case (blas_real)              : return S->is_real();
        case (blas_complex)           : return S->is_complex();


        case (blas_lower_triangular)  : return S->is_lower_triangular(); 
        case (blas_upper_triangular)  : return S->is_lower_triangular(); 

        case (blas_upper_symmetric)   : return S->is_upper_symmetric();
        case (blas_lower_symmetric)   : return S->is_lower_symmetric();

        case (blas_upper_hermitian)   : return S->is_upper_hermitian();
        case (blas_lower_hermitian)   : return S->is_lower_hermitian();



        default:                return -1;  /* invalid property */
    }
} 

/* -------- */
/*  USDS()  */
/* -------- */
int BLAS_usds(int A)
{
  Sp_mat *S  = Table[A];

  S->destroy();
  Table_remove(A);

  return 0;
}

