#ifndef NIST_BLAS_SPARSE_CPP_H
#define NIST_BLAS_SPARSE_CPP_H

#include "blas_enum.h"

#include <iostream>
#include <vector>
#include <complex>
#include <utility>
  /* pair defined here */


#ifdef SPBLAS_ERROR_FATAL
#include <assert>
#define SPBLAS_ASSERT_RETURN(x, ret_val) assert(x)
#else
#define SPBLAS_ASSERT_RETURN(x, ret_val) {if (!(x)) return ret_val;}
#endif

using namespace std;

namespace gov_nist_math_SPBLAS
{




/*
* Level 1 routines
*
*/

template <class T, class const_T>
void BLAS_xusdot( enum blas_conj_type conj_flag, int nz,
    const_T *x,  const int *index,  const_T *y, int incy,
    T *r, enum blas_base_type index_base)
{

  int i=0;
  T t(0);

  if (index_base == blas_one_base)
    y -= incy;

  if (conj_flag == blas_no_conj)
  {
    for (i=0; i<nz; i++)
      t += x[i] * y[index[i]*incy];
  }
  else
    for (i=0; i<nz; i++)
      t += conj(x[i]) * y[index[i]*incy];


  *r = t;
}


template <class T, class const_T>
void BLAS_xusaxpy(int nz, T alpha, const_T *x,
    const int *index, T *y, int incy,
    enum blas_base_type index_base)
{
  int i=0;

  if (index_base == blas_one_base)
    y -= incy;

  for (i=0; i<nz; i++)
    y[index[i]*incy] += alpha * x[i];
}

template <class T, class const_T>
void BLAS_xusga( int nz, const_T *y, int incy, T *x, const int *indx,
              enum blas_base_type index_base )
{
  int i=0;

  if (index_base == blas_one_base)
    y -= incy;

  for (i=0; i<nz; i++)
    x[i] = y[indx[i]*incy];

}

template <class T>
void BLAS_xusgz( int nz, T *y, int incy, T *x, const int *indx,
              enum blas_base_type index_base )
{
  int i=0;

  if (index_base == blas_one_base)
    y -= incy;

  for (i=0; i<nz; i++)
  {
    x[i] = y[indx[i]*incy];
    y[indx[i]*incy] = 0.0;
  }

}


template <class T, class const_T>
void BLAS_xussc(int nz, const_T *x, T *y, int incy, const int *index,
    enum blas_base_type index_base)
{
  int i=0;

  if (index_base == blas_one_base)
    y -= incy;

  for (i=0; i<nz; i++)
    y[index[i]*incy] = x[i];

}

/* end of Level 1 routines */



class Sp_mat
{
  private:

    int num_rows_;
    int num_cols_;
    int num_nonzeros_;

    /* ... */

    int void_;
    int nnew_;
    int open_;
    int valid_;

    int unit_diag_ ;
    int complex_;
    int real_;
    int single_precision_;
    int double_precision_;
    int upper_triangular_;
    int lower_triangular_;
    int upper_symmetric_;
    int lower_symmetric_;
    int upper_hermitian_;
    int lower_hermitian_;
    int general_;

  public:

    Sp_mat(int M, int N) : 
      num_rows_(M),         /* default construction */
      num_cols_(N),
      num_nonzeros_(0),

      void_(0),
      nnew_(1),
      open_(0),
      valid_(0),

      unit_diag_(1),
      complex_(0),
      real_(0),
      single_precision_(0),
      double_precision_(0),
      upper_triangular_(0),
      lower_triangular_(0),
      upper_symmetric_(0),
      lower_symmetric_(0),
      upper_hermitian_(0),
      lower_hermitian_(0),
      general_(0)
      {}


    int& num_rows()           { return num_rows_; }
    int& num_cols()           { return num_cols_; }
    int& num_nonzeros()         { return num_nonzeros_;}

    int num_rows() const        { return num_rows_; }
    int num_cols() const        { return num_cols_; }
    int num_nonzeros() const      { return num_nonzeros_;}

    int is_void() const         { return void_; }
    int is_new() const          { return nnew_; }
    int is_open() const         { return open_; }
    int is_valid() const        { return valid_; }

    int is_unit_diag() const    { return unit_diag_; }
    int is_complex() const        { return complex_;}
    int is_real() const         { return real_;}
    int is_single_precision() const   { return single_precision_;}
    int is_double_precision() const   { return double_precision_;}
    int is_upper_triangular() const   { return upper_triangular_;}
    int is_lower_triangular() const   { return lower_triangular_;}
    int is_triangular() const     { return upper_triangular_ ||
                           lower_triangular_; }

    int is_lower_symmetric() const    { return lower_symmetric_; }
    int is_upper_symmetric() const    { return upper_symmetric_; }
    int is_symmetric() const      { return upper_symmetric_ ||
                           lower_symmetric_; }

    int is_lower_hermitian() const    { return lower_hermitian_; }
    int is_upper_hermitian() const    { return upper_hermitian_; }

    void set_void()       { void_ = 1;  nnew_ = open_ =  valid_ = 0;}
    void set_new()        { nnew_ = 1;  void_ = open_ =  valid_ = 0;}
    void set_open()       { open_ = 1;  void_ = nnew_  = valid_ = 0;}
    void set_valid()      { valid_ = 1; void_ = nnew_ =  open_ = 0; }


    void set_unit_diag()    { unit_diag_ = 1;}
    void set_complex()        {complex_ = 1; }
    void set_real()         { real_ = 1; }
    void set_single_precision()   { single_precision_ = 1; }
    void set_double_precision()   { double_precision_ = 1; }
    void set_upper_triangular()   { upper_triangular_ = 1; }
    void set_lower_triangular()   { lower_triangular_ = 1; }
    void set_upper_symmetric()  { upper_symmetric_ = 1; }
    void set_lower_symmetric()  { lower_symmetric_ = 1; }
    void set_upper_hermitian()  { upper_hermitian_ = 1; }
    void set_lower_hermitian()  { lower_hermitian_ = 1; }

    virtual int end_construction()
    {
      if (is_open() || is_new())
      {
        void_ = 0;
        nnew_ = 0;
        open_ = 0;
        valid_ = 1;

        return 0;
      }
      else
        return 1;
    }
    virtual void print() const;

    virtual void destroy() {};

    virtual ~Sp_mat() {};

};


template <class T>
class TSp_mat : public Sp_mat
{
  private:
    vector< vector< pair<T, int> > > S;
    vector<T> diag;                 /* optional diag if matrix is
                        triangular. Created
                        at end_construction() phase */
  private:


    inline T sp_dot_product( const vector< pair<T, int> > &r, 
        const T* x, int incx ) const
    {
        T sum(0);

        if (incx == 1)
        {
          for ( typename vector< pair<T,int> >::const_iterator p = r.begin(); 
            p < r.end(); p++)
          {
            //sum = sum + p->first * x[p->second];
            sum += p->first * x[p->second];
          }
        }
        else /* incx != 1 */
        {
          for ( typename vector< pair<T,int> >::const_iterator p = r.begin(); 
            p < r.end(); p++)
           {
            //sum = sum + p->first * x[p->second * incx];
            sum += p->first * x[p->second * incx];
            }
        }
    
        return sum;
  }

  inline void sp_axpy( const T& alpha, const vector< pair<T,int> > &r, 
      T*  y, int incy) const
  {
    if (incy == 1)
    {
      for (typename vector< pair<T,int> >::const_iterator p = r.begin(); 
          p < r.end(); p++)
       y[p->second] += alpha * p->first;  

    }
    else /* incy != 1 */
    {
    for (typename vector< pair<T,int> >::const_iterator p = r.begin(); 
        p < r.end(); p++)
      y[incy * p->second] += alpha * p->first;  
    } 
  } 

  void mult_diag(const T& alpha, const T* x, int incx, T* y, int incy) 
      const
  {
    const T* X = x;
    T* Y = y;
    typename vector<T>::const_iterator d= diag.begin();
    for ( ; d < diag.end(); X+=incx, d++, Y+=incy)
    {
      *Y += alpha * *d * *X;
    }
  }


  void nondiag_mult_vec(const T& alpha, const T* x, int incx, 
      T* y, int incy) const
  {

    int M = num_rows();

    if (incy == 1)
    {
      for (int i=0; i<M; i++)
        y[i] += alpha * sp_dot_product(S[i], x, incx);
    }
    else
    {
      for (int i=0; i<M; i++)
        y[i * incy] += alpha * sp_dot_product(S[i], x, incx);
    }
  }

  void nondiag_mult_vec_transpose(const T& alpha, const T* x, int incx, 
      T* y, int incy) const
  {
    /* saxpy: y += (alpha * x[i]) row[i]  */

    int M = num_rows();
    const T* X = x;
    for (int i=0; i<M; i++, X += incx)
      sp_axpy( alpha * *X, S[i], y, incy);
  }

  void mult_vec(const T& alpha, const T* x, int incx, T* y, int incy) 
      const
  {
    nondiag_mult_vec(alpha, x, incx, y, incy);

    if (is_triangular() || is_symmetric())
      mult_diag(alpha, x, incx, y, incy);

    if (is_symmetric())
      nondiag_mult_vec_transpose(alpha, x, incx, y, incy);
  }


  void mult_vec_transpose(const T& alpha, const T* x, int incx, T* y, 
      int incy) const
  {

    nondiag_mult_vec_transpose(alpha, x, incx, y, incy);

    if (is_triangular() || is_symmetric())
      mult_diag(alpha, x, incx, y, incy);

    if (is_symmetric())
      nondiag_mult_vec(alpha, x, incx, y, incy);
  }






      

  int triangular_solve(T alpha, T* x, int incx ) const
  {
    if (alpha == 0.0)
      return -1;

    if ( ! is_triangular() )
      return -1;

    int N = num_rows();

    if (is_lower_triangular())
    {
        for (int i=0, ii=0; i<N; i++, ii += incx)
            x[ii] = (x[ii] - sp_dot_product(S[i], x, incx)) / diag[i];
       if (alpha != 1.0)
       {
        for (int i=0, ii=0; i<N; i++, ii += incx)
            x[ii] /= alpha; 
       }
    }
    else if (is_upper_triangular())
    {
        for (int i=N-1, ii=(N-1)*incx ;   0<=i ;    i--, ii-=incx)
        {
          x[ii] = (x[ii] - sp_dot_product(S[i],x, incx)) / diag[i];
        }
      if (alpha != 1.0)
      {

        for (int i=N-1, ii=(N-1)*incx ;   0<=i ;    i--, ii-=incx)

          x[ii] /= alpha; 
      }

    }
    else
        return 1;

    return 0;
  }

  int transpose_triangular_solve(T alpha, T* x, int incx) const
  {
    if ( ! is_triangular())
      return -1;

    int N = num_rows();

    if (is_lower_triangular())
    {

      for (int j=N-1, jj=(N-1)*incx; 0<=j; j--, jj -= incx)
      {
        x[jj] /= diag[j] ;
        sp_axpy( -x[jj], S[j], x, incx);
      }
      if (alpha != 1.0)
      {
        for (int jj=(N-1)*incx; 0<=jj; jj -=incx)
          x[jj] /= alpha;
      }
    }
    else if (is_upper_triangular())
    {
      
      for (int j=0, jj=0; j<N; j++, jj += incx)
      {
        x[jj] /= diag[j];
        sp_axpy(- x[jj], S[j], x, incx);
      }
      if (alpha != 1.0)
      {
        for (int jj=(N-1)*incx; 0<=jj; jj -=incx)
          x[jj] /= alpha;
      }
    }
    else
         return 1;

    return 0;


  }

 public:

  inline T& val(pair<T, int> &VP) { return VP.first; }
  inline int& col_index(pair<T,int> &VP) { return VP.second; } 

  inline const T& val(pair<T, int> const &VP) const { return VP.first; }
  inline int col_index(pair<T,int> const &VP) const { return VP.second; } 


  TSp_mat( int M, int N) : Sp_mat(M,N), S(M), diag() {}

  void destroy()
  {
    // set vector sizes to zero
    (vector<T>(0)).swap(diag);
    (vector< vector< pair<T, int> > > (0) ).swap(S);
  }

  int insert_entry(T val, int i, int j)
  {
    int triangular = is_upper_triangular() || is_lower_triangular();

    if (is_new())
    {
      set_open();
      
      if (triangular)
      {
        diag.resize(num_rows());
        /* cerr<< "reserved diag.\n"; */
        for (unsigned int ii=0; ii< diag.size(); ii++)
          diag[ii] = (T) 0; 
      }

    }
      if (is_open())
    {

      if (triangular && i==j)
      {
        diag[i] += val;
        /*cerr << "diag["<<i<<"] = " << val << "\n"; */
      }
      else
      {
        S[i].push_back( make_pair(val, j) );
        /* cerr << "push_back: ["<<i <<","<<j<<"] =  " << val <<"\n"; */
      }

      num_nonzeros() ++;
    }


    return 0;
  }

  int insert_entries( int nz, const T* Val, const int *I, const int *J)
  {
    for (int i=0; i<nz; i++)
      insert_entry(Val[i], I[i], J[i]) ;
    return 0;

  }

  void insert_row(int k, int nz, const T* Val, int *J)
  {
    for (int i=0; i<nz; i++)
      insert_entry(Val[i], k, J[i]);  
    return 0;
  }

  void insert_col(int k, int nz, const T* Val, int *I)
  {
    for (int i=0; i<nz; i++)
      insert_entry(Val[i], I[i], k);  
    return 0;
  }

  /* NOTE: I and J are GLOBAL indices, not block offsets like SPBLAS. */
  /* */
  void insert_block(int Mb, int Nb, const T* Val, int row_stride, 
        int col_stride, int I, int J)
  {
    for (i=0; i<Mb; i++)
      for (j=0; j<Nb; j++)
        insert_entry( Val[row_stride*i + col_stride*j], I+i, J+j );
  }

  int end_construction()
  {
    return Sp_mat::end_construction();
  }



  int usmv(enum blas_trans_type transa, T alpha, const  T* x , int incx, 
    T* y, int incy) const
  {
    
  SPBLAS_ASSERT_RETURN(is_valid(), -1);

  if (transa == blas_no_trans)
    mult_vec(alpha, x, incx, y, incy);
  else
    mult_vec_transpose(alpha, x, incx, y, incy);
    
  
    return 0;

  }


  int usmm(blas_order_type ordera, /* enum */blas_trans_type transa, 
    int nrhs, const T& alpha, const  T* b, int ldb, T* C, int ldC) const
  {
    if (ordera == blas_colmajor)
    {
      /* for each column of C, perform a mat_vec */
      for (int i=0; i<nrhs; i++)
        usmv( transa, alpha, b[i], ldb, C[i], ldC );
      return 0;
    }
    else
    if (ordera == blas_rowmajor)
    {
      /* for each column of C, perform a mat_vec */
      for (int i=0; i<nrhs; i++)
        usmv( transa, alpha, b[i*ldb], 1, C[i*ldC], 1 );
      return 0;
    }
    else
      return 1;
  }

  int ussv( enum blas_trans_type transa, const T& alpha,  T* x, int incx) const
  {
      if (transa == blas_trans)
        return transpose_triangular_solve(alpha, x, incx);
      else
        return triangular_solve(alpha, x, incx);


  }


  void print() const
  {
    Sp_mat::print();  /* print matrix header info */

    /* if there is actual data, print out contents */
    for (int i=0; i<num_rows(); i++)
      for (unsigned int j=0; j< S[i].size(); j++)
        cout << i << "    " << col_index(S[i][j]) <<
              "        "  << val(S[i][j]) << "\n";

    /* if matrix is triangular, print out diagonals */
    if (is_upper_triangular() || is_lower_triangular())
    {
      for (unsigned int i=0; i< diag.size(); i++)
        cout << i << "    " << i << "     " << diag[i] << "\n";
    }
  }

    
};



template <class T, class const_T>
int BLAS_xusmv(enum blas_trans_type transa, const T& alpha, int A, 
  const_T *x, int incx, T *y, int incy )
{
  TSp_mat<T> *M = 
        (TSp_mat<T> *) Table[A];

  if (!(M->is_valid()))
    return 1;

  return M->usmv(transa, alpha, x, incx, y, incy);
}




template <class T>
int BLAS_xussv(enum blas_trans_type transa, const T& alpha, int A, 
  T *x, int incx)
{
  TSp_mat<T> *M = 
        (TSp_mat<T> *) Table[A];

  if (!(M->is_valid()))
    return 1;

  return M->ussv(transa, alpha, x, incx);
}


typedef TSp_mat<float> FSp_mat;
typedef TSp_mat<double> DSp_mat;
typedef TSp_mat<complex<float> > CSp_mat;
typedef TSp_mat<complex<double> > ZSp_mat;



void table_print();
void print(int A);

}
/* namespace */


#endif
